"use strict";
exports.__esModule = true;
var fs_extra_1 = require("fs-extra");
var path_1 = require("path");
var isValidPath = require('is-valid-path');
var recurFs = require('recur-fs');
function keyFileBasic(kfsPath, cache) {
    kfsPath = kfsPath || __dirname; // Current working folder by default.
    kfsPath = String(kfsPath);
    if (!isValidPath(kfsPath)) {
        throw new Error('Invalid stroage path.');
    }
    return {
        // Synchronous
        setSync: setSync,
        getSync: getSync,
        deleteSync: deleteSync,
        clearSync: clearSync,
        hasSync: hasSync,
        // Asynchronous
        setAsync: setAsync,
        getAsync: getAsync,
        deleteAsync: deleteAsync,
        clearAsync: clearAsync,
        hasAsync: hasAsync,
        // Iterate
        querySync: querySync,
        queryAsync: queryAsync
    };
    function setSync(key, value) {
        if (value === undefined) {
            return deleteSync(key);
        }
        key = validizeKey(key);
        var file = path_1.join(kfsPath, key);
        fs_extra_1.outputJsonSync(file, value);
        return (cache[key] = value);
    }
    function getSync(key) {
        key = validizeKey(key);
        if (key in cache) {
            return cache[key];
        }
        var file = path_1.join(kfsPath, key);
        try {
            var status = fs_extra_1.statSync(file);
            if (!status || !status.isFile()) {
                return (cache[key] = null);
            }
        }
        catch (err) {
            return (cache[key] = null);
        }
        return (cache[key] = fs_extra_1.readJsonSync(file));
    }
    function deleteSync(key) {
        key = validizeKey(key);
        if (key === '*') {
            return clearSync();
        }
        var file = path_1.join(kfsPath, key);
        fs_extra_1.removeSync(file);
        return delete cache[key];
    }
    function clearSync() {
        fs_extra_1.removeSync(kfsPath);
        if (cache.constructor === Object) {
            cache = {
            /*NEW-EMPTY-CACHE*/
            };
            return true;
        }
        else {
            return delete cache['*'];
        }
    }
    function hasSync(key) {
        key = validizeKey(key);
        if (key in cache) {
            return true;
        }
        var file = path_1.join(kfsPath, key);
        try {
            var status = fs_extra_1.statSync(file);
            if (!status || !status.isFile()) {
                return false;
            }
        }
        catch (err) {
            return false;
        }
        return true;
    }
    function setAsync(key, value) {
        if (value === undefined) {
            return deleteAsync(key);
        }
        key = validizeKey(key);
        var file = path_1.join(kfsPath, key);
        return new Promise(function (resolve, reject) {
            fs_extra_1.outputJson(file, value, function (err) {
                if (err) {
                    reject(err);
                }
                else {
                    resolve((cache[key] = value));
                }
            });
        });
    }
    function getAsync(key) {
        key = validizeKey(key);
        if (key in cache) {
            return Promise.resolve(cache[key]);
        }
        else {
            var file = path_1.join(kfsPath, key);
            return new Promise(function (resolve, reject) {
                fs_extra_1.stat(file, function (err, status) {
                    if (err || !status || !status.isFile()) {
                        resolve((cache[key] = null));
                    }
                    else {
                        fs_extra_1.readJson(file, function (err, value) {
                            if (err) {
                                reject(err);
                            }
                            else {
                                resolve((cache[key] = value));
                            }
                        });
                    }
                });
            });
        }
    }
    function deleteAsync(key) {
        key = validizeKey(key);
        if (key === '*') {
            return clearAsync();
        }
        var file = path_1.join(kfsPath, key);
        return new Promise(function (resolve, reject) {
            fs_extra_1.remove(file, function (err) {
                if (err) {
                    reject(err);
                }
                else {
                    resolve(delete cache[key]);
                }
            });
        });
    }
    function clearAsync() {
        return new Promise(function (resolve, reject) {
            fs_extra_1.remove(kfsPath, function (err) {
                if (err) {
                    reject(err);
                }
                else {
                    if (cache.constructor === Object) {
                        cache = {
                        /*NEW-EMPTY-CACHE*/
                        };
                        resolve(true);
                    }
                    else {
                        resolve(delete cache['*']);
                    }
                }
            });
        });
    }
    function hasAsync(key) {
        key = validizeKey(key);
        if (key in cache) {
            return Promise.resolve(true);
        }
        else {
            var file = path_1.join(kfsPath, key);
            return new Promise(function (resolve, reject) {
                fs_extra_1.stat(file, function (err, status) {
                    if (err || !status || !status.isFile()) {
                        resolve(false);
                    }
                    else {
                        resolve(true);
                    }
                });
            });
        }
    }
    function querySync(collection) {
        collection = path_1.join(kfsPath, validizeKey(collection));
        if (collection in cache)
            return cache[collection];
        try {
            var files = recurFs.readdir.sync(collection, function (resource, status) {
                return status.isFile();
            });
            files = files.map(function (file) { return path_1.relative(kfsPath, file); });
            return (cache[collection] = files || []);
        }
        catch (err) {
            return [];
        }
    }
    function queryAsync(collection) {
        collection = path_1.join(kfsPath, validizeKey(collection));
        if (collection in cache)
            return Promise.resolve(cache[collection]);
        return new Promise(function (resolve, reject) {
            //// This implementation does not work with empty folders:
            // recurFs.readdir(collection, function(resource, status, next) {
            //     next(status.isFile());
            // }, function(err, resources) {
            //     if (err) {
            //         reject(err);
            //     }
            //     else {
            //         resolve(resources.map(file => path.relative(kfsPath, file)));
            //     }
            // });
            var fileList = [], jobNumber = 1, terminated = false;
            fs_extra_1.stat(collection, function (err, status) {
                if (err) {
                    if (err.code === 'ENOENT')
                        resolve((cache[collection] = []));
                    else
                        reject(err);
                }
                else {
                    processFolder(collection);
                }
            });
            function processFolder(folder) {
                if (terminated)
                    return;
                fs_extra_1.readdir(folder, function (err, files) {
                    if (terminated)
                        return;
                    jobNumber--;
                    if (err) {
                        terminated = true;
                        reject(err);
                    }
                    jobNumber += files.length;
                    if (!jobNumber) {
                        resolve((cache[collection] = fileList));
                    }
                    files.forEach(function (file) {
                        if (terminated)
                            return;
                        var filePath = path_1.join(folder, file);
                        fs_extra_1.stat(filePath, function (err, status) {
                            if (terminated)
                                return;
                            jobNumber--;
                            if (err) {
                                terminated = true;
                                reject(err);
                            }
                            if (status.isFile()) {
                                fileList.push(path_1.relative(kfsPath, filePath));
                            }
                            else if (status.isDirectory()) {
                                jobNumber++;
                                processFolder(filePath);
                            }
                            if (!jobNumber) {
                                resolve((cache[collection] = fileList));
                            }
                        });
                    });
                });
            }
        });
    }
    ///////////////////////////////////////////////////
    function validizeKey(key) {
        key = String(key);
        if (key.indexOf('/..') >= 0 || key.indexOf('../') >= 0 || key === '..') {
            throw new Error('Invalid key name.');
        }
        return key;
    }
}
exports["default"] = keyFileBasic;
